@@ -1,12 +1,12 @@
 /////////////////////////////////////////////////////////////////////////////
-// Name:        log.cpp
+// Name:        src/common/log.cpp
 // Purpose:     Assorted wxLogXXX functions, and wxLog (sink for logs)
 // Author:      Vadim Zeitlin
 // Modified by:
 // Created:     29/01/98
-// RCS-ID:      $Id: log.cpp,v 1.128.2.2 2003/01/01 05:04:49 RD Exp $
+// RCS-ID:      $Id: log.cpp 50994 2008-01-02 21:27:31Z VZ $
 // Copyright:   (c) 1998 Vadim Zeitlin <zeitlin@dptmaths.ens-cachan.fr>
-// Licence:     wxWindows license
+// Licence:     wxWindows licence
 /////////////////////////////////////////////////////////////////////////////
 
 // ============================================================================
@@ -17,52 +17,48 @@
 // headers
 // ----------------------------------------------------------------------------
 
-#ifdef __GNUG__
-  #pragma implementation "log.h"
-#endif
-
 // For compilers that support precompilation, includes "wx.h".
 #include "wx/wxprec.h"
 
 #ifdef __BORLANDC__
-  #pragma hdrstop
+    #pragma hdrstop
 #endif
 
-// wxWindows
-#ifndef WX_PRECOMP
-    #include "wx/string.h"
-    #include "wx/intl.h"
-    #include "wx/app.h"
-
-    #if wxUSE_GUI
-        #include "wx/window.h"
-        #include "wx/msgdlg.h"
-        #ifdef __WXMSW__
-            #include "wx/msw/private.h"
-        #endif
-    #endif
-#endif //WX_PRECOMP
-
-#include  "wx/file.h"
-#include  "wx/textfile.h"
-#include  "wx/utils.h"
-#include  "wx/wxchar.h"
-#include  "wx/log.h"
-#include  "wx/thread.h"
-
 #if wxUSE_LOG
 
-// other standard headers
-#include  <errno.h>
-#include  <stdlib.h>
-#include  <time.h>
+// wxWidgets
+#ifndef WX_PRECOMP
+    #include "wx/log.h"
+    #include "wx/app.h"
+    #include "wx/arrstr.h"
+    #include "wx/intl.h"
+    #include "wx/string.h"
+    #include "wx/utils.h"
+#endif //WX_PRECOMP
 
-#if defined(__WXMSW__)
-  #include  "wx/msw/private.h"      // includes windows.h for OutputDebugString
+#include "wx/apptrait.h"
+#include "wx/datetime.h"
+#include "wx/file.h"
+#include "wx/msgout.h"
+#include "wx/textfile.h"
+#include "wx/thread.h"
+#include "wx/wxchar.h"
+
+// other standard headers
+#ifndef __WXWINCE__
+#include <errno.h>
 #endif
 
-#if defined(__WXMAC__)
-  #include  "wx/mac/private.h"  // includes mac headers
+#include <stdlib.h>
+
+#ifndef __WXWINCE__
+#include <time.h>
+#else
+#include "wx/msw/wince/time.h"
+#endif
+
+#if defined(__WINDOWS__)
+    #include "wx/msw/private.h" // includes windows.h
 #endif
 
 // ----------------------------------------------------------------------------
@@ -80,33 +76,6 @@
 // implementation
 // ============================================================================
 
-// ----------------------------------------------------------------------------
-// globals
-// ----------------------------------------------------------------------------
-
-// log functions can't allocate memory (LogError("out of memory...") should
-// work!), so we use a static buffer for all log messages
-#define LOG_BUFFER_SIZE   (4096)
-
-// static buffer for error messages
-static wxChar   s_szBufStatic[LOG_BUFFER_SIZE];
-
-static wxChar  *s_szBuf     = s_szBufStatic;
-static size_t   s_szBufSize = WXSIZEOF( s_szBufStatic );
-
-#if wxUSE_THREADS
-
-// the critical section protecting the static buffer
-static wxCriticalSection gs_csLogBuf;
-
-#endif // wxUSE_THREADS
-
-// return true if we have a non NULL non disabled log target
-static inline bool IsLoggingEnabled()
-{
-    return wxLog::IsEnabled() && (wxLog::GetActiveTarget() != NULL);
-}
-
 // ----------------------------------------------------------------------------
 // implementation of Log functions
 //
@@ -117,12 +86,8 @@ static inline bool IsLoggingEnabled()
 // generic log function
 void wxVLogGeneric(wxLogLevel level, const wxChar *szFormat, va_list argptr)
 {
-    if ( IsLoggingEnabled() ) {
-        wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);
-
-        wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);
-
-        wxLog::OnLog(level, s_szBuf, time(NULL));
+    if ( wxLog::IsEnabled() ) {
+        wxLog::OnLog(level, wxString::FormatV(szFormat, argptr), time(NULL));
     }
 }
 
@@ -137,14 +102,12 @@ void wxLogGeneric(wxLogLevel level, const wxChar *szFormat, ...)
 #define IMPLEMENT_LOG_FUNCTION(level)                               \
   void wxVLog##level(const wxChar *szFormat, va_list argptr)        \
   {                                                                 \
-    if ( IsLoggingEnabled() ) {                                     \
-      wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);                      \
-                                                                    \
-      wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);    \
-                                                                    \
-      wxLog::OnLog(wxLOG_##level, s_szBuf, time(NULL));             \
+    if ( wxLog::IsEnabled() ) {                                     \
+      wxLog::OnLog(wxLOG_##level,                                   \
+                   wxString::FormatV(szFormat, argptr), time(NULL));\
     }                                                               \
   }                                                                 \
+                                                                    \
   void wxLog##level(const wxChar *szFormat, ...)                    \
   {                                                                 \
     va_list argptr;                                                 \
@@ -165,6 +128,7 @@ void wxSafeShowMessage(const wxString& title, const wxString& text)
     ::MessageBox(NULL, text, title, MB_OK | MB_ICONSTOP);
 #else
     wxFprintf(stderr, _T("%s: %s\n"), title.c_str(), text.c_str());
+    fflush(stderr);
 #endif
 }
 
@@ -172,11 +136,13 @@ void wxSafeShowMessage(const wxString& title, const wxString& text)
 // always terminate the program
 void wxVLogFatalError(const wxChar *szFormat, va_list argptr)
 {
-    wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);
-
-    wxSafeShowMessage(_T("Fatal Error"), s_szBuf);
+    wxSafeShowMessage(_T("Fatal Error"), wxString::FormatV(szFormat, argptr));
 
+#ifdef __WXWINCE__
+    ExitThread(3);
+#else
     abort();
+#endif
 }
 
 void wxLogFatalError(const wxChar *szFormat, ...)
@@ -184,20 +150,19 @@ void wxLogFatalError(const wxChar *szFormat, ...)
     va_list argptr;
     va_start(argptr, szFormat);
     wxVLogFatalError(szFormat, argptr);
-    va_end(argptr);
+
+    // some compilers warn about unreachable code and it shouldn't matter
+    // for the others anyhow...
+    //va_end(argptr);
 }
 
 // same as info, but only if 'verbose' mode is on
 void wxVLogVerbose(const wxChar *szFormat, va_list argptr)
 {
-    if ( IsLoggingEnabled() ) {
-        wxLog *pLog = wxLog::GetActiveTarget();
-        if ( pLog != NULL && pLog->GetVerbose() ) {
-            wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);
-
-            wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);
-
-            wxLog::OnLog(wxLOG_Info, s_szBuf, time(NULL));
+    if ( wxLog::IsEnabled() ) {
+        if ( wxLog::GetActiveTarget() != NULL && wxLog::GetVerbose() ) {
+            wxLog::OnLog(wxLOG_Info,
+                         wxString::FormatV(szFormat, argptr), time(NULL));
         }
     }
 }
@@ -215,14 +180,12 @@ void wxLogVerbose(const wxChar *szFormat, ...)
 #define IMPLEMENT_LOG_DEBUG_FUNCTION(level)                         \
   void wxVLog##level(const wxChar *szFormat, va_list argptr)        \
   {                                                                 \
-    if ( IsLoggingEnabled() ) {                                     \
-      wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);                      \
-                                                                    \
-      wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);    \
-                                                                    \
-      wxLog::OnLog(wxLOG_##level, s_szBuf, time(NULL));             \
+    if ( wxLog::IsEnabled() ) {                                     \
+      wxLog::OnLog(wxLOG_##level,                                   \
+                   wxString::FormatV(szFormat, argptr), time(NULL));\
     }                                                               \
   }                                                                 \
+                                                                    \
   void wxLog##level(const wxChar *szFormat, ...)                    \
   {                                                                 \
     va_list argptr;                                                 \
@@ -233,26 +196,11 @@ void wxLogVerbose(const wxChar *szFormat, ...)
 
   void wxVLogTrace(const wxChar *mask, const wxChar *szFormat, va_list argptr)
   {
-    if ( IsLoggingEnabled() && wxLog::IsAllowedTraceMask(mask) ) {
-      wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);
-
-      wxChar *p = s_szBuf;
-      size_t len = s_szBufSize;
-      wxStrncpy(s_szBuf, _T("("), len);
-      len -= 1; // strlen("(")
-      p += 1;
-      wxStrncat(p, mask, len);
-      size_t lenMask = wxStrlen(mask);
-      len -= lenMask;
-      p += lenMask;
-
-      wxStrncat(p, _T(") "), len);
-      len -= 2;
-      p += 2;
-
-      wxVsnprintf(p, len, szFormat, argptr);
-
-      wxLog::OnLog(wxLOG_Trace, s_szBuf, time(NULL));
+    if ( wxLog::IsEnabled() && wxLog::IsAllowedTraceMask(mask) ) {
+      wxString msg;
+      msg << _T("(") << mask << _T(") ") << wxString::FormatV(szFormat, argptr);
+          
+      wxLog::OnLog(wxLOG_Trace, msg, time(NULL));
     }
   }
 
@@ -269,12 +217,8 @@ void wxLogVerbose(const wxChar *szFormat, ...)
     // we check that all of mask bits are set in the current mask, so
     // that wxLogTrace(wxTraceRefCount | wxTraceOle) will only do something
     // if both bits are set.
-    if ( IsLoggingEnabled() && ((wxLog::GetTraceMask() & mask) == mask) ) {
-      wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);
-
-      wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);
-
-      wxLog::OnLog(wxLOG_Trace, s_szBuf, time(NULL));
+    if ( wxLog::IsEnabled() && ((wxLog::GetTraceMask() & mask) == mask) ) {
+      wxLog::OnLog(wxLOG_Trace, wxString::FormatV(szFormat, argptr), time(NULL));
     }
   }
 
@@ -296,26 +240,15 @@ IMPLEMENT_LOG_DEBUG_FUNCTION(Trace)
 // wxLogSysError: one uses the last error code, for other  you must give it
 // explicitly
 
-// common part of both wxLogSysError
-void wxLogSysErrorHelper(long lErrCode)
+// return the system error message description
+static inline wxString wxLogSysErrorHelper(long err)
 {
-    wxChar szErrMsg[LOG_BUFFER_SIZE / 2];
-    wxSnprintf(szErrMsg, WXSIZEOF(szErrMsg),
-               _(" (error %ld: %s)"), lErrCode, wxSysErrorMsg(lErrCode));
-    wxStrncat(s_szBuf, szErrMsg, s_szBufSize - wxStrlen(s_szBuf));
-
-    wxLog::OnLog(wxLOG_Error, s_szBuf, time(NULL));
+    return wxString::Format(_(" (error %ld: %s)"), err, wxSysErrorMsg(err));
 }
 
 void WXDLLEXPORT wxVLogSysError(const wxChar *szFormat, va_list argptr)
 {
-    if ( IsLoggingEnabled() ) {
-        wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);
-
-        wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);
-
-        wxLogSysErrorHelper(wxSysErrorCode());
-    }
+    wxVLogSysError(wxSysErrorCode(), szFormat, argptr);
 }
 
 void WXDLLEXPORT wxLogSysError(const wxChar *szFormat, ...)
@@ -326,14 +259,12 @@ void WXDLLEXPORT wxLogSysError(const wxChar *szFormat, ...)
     va_end(argptr);
 }
 
-void WXDLLEXPORT wxVLogSysError(long lErrCode, const wxChar *szFormat, va_list argptr)
+void WXDLLEXPORT wxVLogSysError(long err, const wxChar *fmt, va_list argptr)
 {
-    if ( IsLoggingEnabled() ) {
-        wxCRIT_SECT_LOCKER(locker, gs_csLogBuf);
-
-        wxVsnprintf(s_szBuf, s_szBufSize, szFormat, argptr);
-
-        wxLogSysErrorHelper(lErrCode);
+    if ( wxLog::IsEnabled() ) {
+        wxLog::OnLog(wxLOG_Error,
+                     wxString::FormatV(fmt, argptr) + wxLogSysErrorHelper(err),
+                     time(NULL));
     }
 }
 
@@ -349,45 +280,108 @@ void WXDLLEXPORT wxLogSysError(long lErrCode, const wxChar *szFormat, ...)
 // wxLog class implementation
 // ----------------------------------------------------------------------------
 
-wxLog::wxLog()
+// define a critical section gs_prevCS protecting access to wxLog::ms_prevXXX
+wxCRIT_SECT_DECLARE(gs_prevCS);
+
+/* static */
+unsigned wxLog::DoLogNumberOfRepeats()
 {
-    m_bHasMessages = FALSE;
+    wxLog * const pLogger = GetActiveTarget();
+    return pLogger ? pLogger->LogLastRepeatIfNeeded() : 0u;
 }
 
-wxChar *wxLog::SetLogBuffer( wxChar *buf, size_t size)
+unsigned wxLog::LogLastRepeatIfNeeded()
 {
-    wxChar *oldbuf = s_szBuf;
+    wxCRIT_SECT_LOCKER(lock, gs_prevCS);
 
-    if( buf == 0 )
-    {
-        s_szBuf = s_szBufStatic;
-        s_szBufSize = WXSIZEOF( s_szBufStatic );
-    }
-    else
-    {
-        s_szBuf = buf;
-        s_szBufSize = size;
-    }
-
-    return (oldbuf == s_szBufStatic ) ? 0 : oldbuf;
+    return LogLastRepeatIfNeededUnlocked();
 }
 
+unsigned wxLog::LogLastRepeatIfNeededUnlocked()
+{
+    long retval = ms_prevCounter;
+    if ( ms_prevCounter > 0 )
+    {
+        wxString msg;
+#if wxUSE_INTL
+        msg.Printf(wxPLURAL("The previous message repeated once.",
+                            "The previous message repeated %lu times.",
+                            ms_prevCounter),
+                   ms_prevCounter);
+#else
+        msg.Printf(wxT("The previous message was repeated."));
+#endif
+        ms_prevCounter = 0;
+        ms_prevString.clear();
+        DoLog(ms_prevLevel, msg.c_str(), ms_prevTimeStamp);
+    }
+    return retval;
+}
+
+wxLog::~wxLog()
+{
+}
+
+/* static */
+void wxLog::OnLog(wxLogLevel level, const wxChar *szString, time_t t)
+{
+    if ( IsEnabled() && ms_logLevel >= level )
+    {
+        wxLog *pLogger = GetActiveTarget();
+        if ( pLogger )
+        {
+            if ( GetRepetitionCounting() )
+            {
+                wxCRIT_SECT_LOCKER(lock, gs_prevCS);
+
+                if ( szString == ms_prevString )
+                {
+                    ms_prevCounter++;
+
+                    // nothing else to do, in particular, don't log the
+                    // repeated message
+                    return;
+                }
+
+                pLogger->LogLastRepeatIfNeededUnlocked();
+
+                // reset repetition counter for a new message
+                ms_prevString = szString;
+                ms_prevLevel = level;
+                ms_prevTimeStamp = t;
+            }
+
+            pLogger->DoLog(level, szString, t);
+        }
+    }
+}
+
+// deprecated function
+#if WXWIN_COMPATIBILITY_2_6
+
+wxChar *wxLog::SetLogBuffer(wxChar * WXUNUSED(buf), size_t WXUNUSED(size))
+{
+    return NULL;
+}
+
+#endif // WXWIN_COMPATIBILITY_2_6
+
 wxLog *wxLog::GetActiveTarget()
 {
     if ( ms_bAutoCreate && ms_pLogger == NULL ) {
         // prevent infinite recursion if someone calls wxLogXXX() from
         // wxApp::CreateLogTarget()
-        static bool s_bInGetActiveTarget = FALSE;
+        static bool s_bInGetActiveTarget = false;
         if ( !s_bInGetActiveTarget ) {
-            s_bInGetActiveTarget = TRUE;
+            s_bInGetActiveTarget = true;
 
             // ask the application to create a log target for us
             if ( wxTheApp != NULL )
-                ms_pLogger = wxTheApp->CreateLogTarget();
+                ms_pLogger = wxTheApp->GetTraits()->CreateLogTarget();
             else
                 ms_pLogger = new wxLogStderr;
 
-            s_bInGetActiveTarget = FALSE;
+            s_bInGetActiveTarget = false;
 
             // do nothing if it fails - what can we do?
         }
@@ -412,7 +406,7 @@ wxLog *wxLog::SetActiveTarget(wxLog *pLogger)
 
 void wxLog::DontCreateOnDemand()
 {
-    ms_bAutoCreate = FALSE;
+    ms_bAutoCreate = false;
 
     // this is usually called at the end of the program and we assume that it
     // is *always* called at the end - so we free memory here to avoid false
@@ -420,11 +414,16 @@ void wxLog::DontCreateOnDemand()
     ClearTraceMasks();
 }
 
+void wxLog::DoCreateOnDemand()
+{
+    ms_bAutoCreate = true;
+}
+
 void wxLog::RemoveTraceMask(const wxString& str)
 {
     int index = ms_aTraceMasks.Index(str);
     if ( index != wxNOT_FOUND )
-        ms_aTraceMasks.Remove((size_t)index);
+        ms_aTraceMasks.RemoveAt((size_t)index);
 }
 
 void wxLog::ClearTraceMasks()
@@ -434,16 +433,21 @@ void wxLog::ClearTraceMasks()
 
 void wxLog::TimeStamp(wxString *str)
 {
+#if wxUSE_DATETIME
     if ( ms_timestamp )
     {
         wxChar buf[256];
         time_t timeNow;
         (void)time(&timeNow);
-        wxStrftime(buf, WXSIZEOF(buf), ms_timestamp, localtime(&timeNow));
+
+        struct tm tm;
+        wxStrftime(buf, WXSIZEOF(buf),
+                    ms_timestamp, wxLocaltime_r(&timeNow, &tm));
 
         str->Empty();
         *str << buf << wxT(": ");
     }
+#endif // wxUSE_DATETIME
 }
 
 void wxLog::DoLog(wxLogLevel level, const wxChar *szString, time_t t)
@@ -453,7 +457,11 @@ void wxLog::DoLog(wxLogLevel level, const wxChar *szString, time_t t)
             DoLogString(wxString(_("Fatal error: ")) + szString, t);
             DoLogString(_("Program aborted."), t);
             Flush();
+#ifdef __WXWINCE__
+            ExitThread(3);
+#else
             abort();
+#endif
             break;
 
         case wxLOG_Error:
@@ -493,8 +501,61 @@ void wxLog::DoLogString(const wxChar *WXUNUSED(szString), time_t WXUNUSED(t))
 
 void wxLog::Flush()
 {
-    // remember that we don't have any more messages to show
-    m_bHasMessages = FALSE;
+    LogLastRepeatIfNeeded();
+}
+
+/*static*/ bool wxLog::IsAllowedTraceMask(const wxChar *mask)
+{
+    for ( wxArrayString::iterator it = ms_aTraceMasks.begin(),
+                                  en = ms_aTraceMasks.end();
+         it != en; ++it )
+        if ( *it == mask)
+            return true;
+    return false;
+}
+
+// ----------------------------------------------------------------------------
+// wxLogBuffer implementation
+// ----------------------------------------------------------------------------
+
+void wxLogBuffer::Flush()
+{
+    if ( !m_str.empty() )
+    {
+        wxMessageOutputBest out;
+        out.Printf(_T("%s"), m_str.c_str());
+        m_str.clear();
+    }
+}
+
+void wxLogBuffer::DoLog(wxLogLevel level, const wxChar *szString, time_t t)
+{
+    switch ( level )
+    {
+        case wxLOG_Trace:
+        case wxLOG_Debug:
+#ifdef __WXDEBUG__
+            // don't put debug messages in the buffer, we don't want to show
+            // them to the user in a msg box, log them immediately
+            {
+                wxString str;
+                TimeStamp(&str);
+                str += szString;
+
+                wxMessageOutputDebug dbgout;
+                dbgout.Printf(_T("%s\n"), str.c_str());
+            }
+#endif // __WXDEBUG__
+            break;
+
+        default:
+            wxLog::DoLog(level, szString, t);
+    }
+}
+
+void wxLogBuffer::DoLogString(const wxChar *szString, time_t WXUNUSED(t))
+{
+    m_str << szString << _T("\n");
 }
 
 // ----------------------------------------------------------------------------
@@ -509,234 +570,29 @@ wxLogStderr::wxLogStderr(FILE *fp)
         m_fp = fp;
 }
 
-#if defined(__WXMAC__) && !defined(__DARWIN__) && defined(__MWERKS__) && (__MWERKS__ >= 0x2400)
-
-// MetroNub stuff doesn't seem to work in CodeWarrior 5.3 Carbon builds...
-
-#ifndef __MetroNubUtils__
-#include "MetroNubUtils.h"
-#endif
-
-#ifdef __cplusplus
-    extern "C" {
-#endif
-
-#ifndef __GESTALT__
-#include <Gestalt.h>
-#endif
-
-#ifndef true
-#define true 1
-#endif
-
-#ifndef false
-#define false 0
-#endif
-
-#if TARGET_API_MAC_CARBON
-
-    #include <CodeFragments.h>
-
-    EXTERN_API_C( long )
-    CallUniversalProc(UniversalProcPtr theProcPtr, ProcInfoType procInfo, ...);
-
-    ProcPtr gCallUniversalProc_Proc = NULL;
-
-#endif
-
-static MetroNubUserEntryBlock*    gMetroNubEntry = NULL;
-
-static long fRunOnce = false;
-
-Boolean IsCompatibleVersion(short inVersion);
-
-/* ---------------------------------------------------------------------------
-        IsCompatibleVersion
-   --------------------------------------------------------------------------- */
-
-Boolean IsCompatibleVersion(short inVersion)
-{
-    Boolean result = false;
-
-    if (fRunOnce)
-    {
-        MetroNubUserEntryBlock* block = (MetroNubUserEntryBlock *)result;
-
-        result = (inVersion <= block->apiHiVersion);
-    }
-
-    return result;
-}
-
-/* ---------------------------------------------------------------------------
-        IsMetroNubInstalled
-   --------------------------------------------------------------------------- */
-
-Boolean IsMetroNubInstalled()
-{
-    if (!fRunOnce)
-    {
-        long result, value;
-
-        fRunOnce = true;
-        gMetroNubEntry = NULL;
-
-        if (Gestalt(gestaltSystemVersion, &value) == noErr && value < 0x1000)
-        {
-            /* look for MetroNub's Gestalt selector */
-            if (Gestalt(kMetroNubUserSignature, &result) == noErr)
-            {
-
-            #if TARGET_API_MAC_CARBON
-                if (gCallUniversalProc_Proc == NULL)
-                {
-                    CFragConnectionID   connectionID;
-                    Ptr                 mainAddress;
-                    Str255              errorString;
-                    ProcPtr             symbolAddress;
-                    OSErr               err;
-                    CFragSymbolClass    symbolClass;
-
-                    symbolAddress = NULL;
-                    err = GetSharedLibrary("\pInterfaceLib", kPowerPCCFragArch, kFindCFrag,
-                                           &connectionID, &mainAddress, errorString);
-
-                    if (err != noErr)
-                    {
-                        gCallUniversalProc_Proc = NULL;
-                        goto end;
-                    }
-
-                    err = FindSymbol(connectionID, "\pCallUniversalProc",
-                                    (Ptr *) &gCallUniversalProc_Proc, &symbolClass);
-
-                    if (err != noErr)
-                    {
-                        gCallUniversalProc_Proc = NULL;
-                        goto end;
-                    }
-                }
-            #endif
-
-                {
-                    MetroNubUserEntryBlock* block = (MetroNubUserEntryBlock *)result;
-
-                    /* make sure the version of the API is compatible */
-                    if (block->apiLowVersion <= kMetroNubUserAPIVersion &&
-                        kMetroNubUserAPIVersion <= block->apiHiVersion)
-                        gMetroNubEntry = block;        /* success! */
-                }
-
-            }
-        }
-    }
-
-end:
-
-#if TARGET_API_MAC_CARBON
-    return (gMetroNubEntry != NULL && gCallUniversalProc_Proc != NULL);
-#else
-    return (gMetroNubEntry != NULL);
-#endif
-}
-
-/* ---------------------------------------------------------------------------
-        IsMWDebuggerRunning                                            [v1 API]
-   --------------------------------------------------------------------------- */
-
-Boolean IsMWDebuggerRunning()
-{
-    if (IsMetroNubInstalled())
-        return CallIsDebuggerRunningProc(gMetroNubEntry->isDebuggerRunning);
-    else
-        return false;
-}
-
-/* ---------------------------------------------------------------------------
-        AmIBeingMWDebugged                                            [v1 API]
-   --------------------------------------------------------------------------- */
-
-Boolean AmIBeingMWDebugged()
-{
-    if (IsMetroNubInstalled())
-        return CallAmIBeingDebuggedProc(gMetroNubEntry->amIBeingDebugged);
-    else
-        return false;
-}
-
-/* ---------------------------------------------------------------------------
-        UserSetWatchPoint                                            [v2 API]
-   --------------------------------------------------------------------------- */
-
-OSErr UserSetWatchPoint (Ptr address, long length, WatchPointIDT* watchPointID)
-{
-    if (IsMetroNubInstalled() && IsCompatibleVersion(kMetroNubUserAPIVersion))
-        return CallUserSetWatchPointProc(gMetroNubEntry->userSetWatchPoint,
-                                         address, length, watchPointID);
-    else
-        return errProcessIsNotClient;
-}
-
-/* ---------------------------------------------------------------------------
-        ClearWatchPoint                                                [v2 API]
-   --------------------------------------------------------------------------- */
-
-OSErr ClearWatchPoint (WatchPointIDT watchPointID)
-{
-    if (IsMetroNubInstalled() && IsCompatibleVersion(kMetroNubUserAPIVersion))
-        return CallClearWatchPointProc(gMetroNubEntry->clearWatchPoint, watchPointID);
-    else
-        return errProcessIsNotClient;
-}
-
-#ifdef __cplusplus
-    }
-#endif
-
-#endif // defined(__WXMAC__) && !defined(__DARWIN__) && (__MWERKS__ >= 0x2400)
-
 void wxLogStderr::DoLogString(const wxChar *szString, time_t WXUNUSED(t))
 {
     wxString str;
     TimeStamp(&str);
     str << szString;
 
-    fputs(str.mb_str(), m_fp);
-    fputc(_T('\n'), m_fp);
+    wxFputs(str, m_fp);
+    wxFputc(_T('\n'), m_fp);
     fflush(m_fp);
 
-    // under Windows, programs usually don't have stderr at all, so show the
-    // messages also under debugger - unless it's a console program
-#if defined(__WXMSW__) && wxUSE_GUI && !defined(__WXMICROWIN__)
-    str += wxT("\r\n") ;
-    OutputDebugString(str.c_str());
-#endif // MSW
-#if defined(__WXMAC__) && !defined(__DARWIN__) && wxUSE_GUI
-    Str255 pstr ;
-    strcpy( (char*) pstr , str.c_str() ) ;
-    strcat( (char*) pstr , ";g" ) ;
-    c2pstr( (char*) pstr ) ;
-
-    Boolean running = false ;
-
-#if defined(__MWERKS__) && (__MWERKS__ >= 0x2400)
-
-    if ( IsMWDebuggerRunning() && AmIBeingMWDebugged() )
+    // under GUI systems such as Windows or Mac, programs usually don't have
+    // stderr at all, so show the messages also somewhere else, typically in
+    // the debugger window so that they go at least somewhere instead of being
+    // simply lost
+    if ( m_fp == stderr )
     {
-        running = true ;
+        wxAppTraits *traits = wxTheApp ? wxTheApp->GetTraits() : NULL;
+        if ( traits && !traits->HasStderr() )
+        {
+            wxMessageOutputDebug dbgout;
+            dbgout.Printf(_T("%s\n"), str.c_str());
+        }
     }
-
-#endif
-
-    if (running)
-    {
-#ifdef __powerc
-        DebugStr(pstr);
-#else
-        SysBreakStr(pstr);
-#endif
-    }
-#endif // Mac
 }
 
 // ----------------------------------------------------------------------------
@@ -744,6 +600,7 @@ void wxLogStderr::DoLogString(const wxChar *szString, time_t WXUNUSED(t))
 // ----------------------------------------------------------------------------
 
 #if wxUSE_STD_IOSTREAM
+#include "wx/ioswrap.h"
 wxLogStream::wxLogStream(wxSTD ostream *ostr)
 {
     if ( ostr == NULL )
@@ -756,7 +613,7 @@ void wxLogStream::DoLogString(const wxChar *szString, time_t WXUNUSED(t))
 {
     wxString str;
     TimeStamp(&str);
-    (*m_ostr) << str << wxConvertWX2MB(szString) << wxSTD endl;
+    (*m_ostr) << wxSafeConvertWX2MB(str) << wxSafeConvertWX2MB(szString) << wxSTD endl;
 }
 #endif // wxUSE_STD_IOSTREAM
 
@@ -766,7 +623,7 @@ void wxLogStream::DoLogString(const wxChar *szString, time_t WXUNUSED(t))
 
 wxLogChain::wxLogChain(wxLog *logger)
 {
-    m_bPassMessages = TRUE;
+    m_bPassMessages = true;
 
     m_logNew = logger;
     m_logOld = wxLog::SetActiveTarget(this);
@@ -793,7 +650,7 @@ void wxLogChain::Flush()
     if ( m_logOld )
         m_logOld->Flush();
 
-    // be careful to avoid inifinite recursion
+    // be careful to avoid infinite recursion
     if ( m_logNew && m_logNew != this )
         m_logNew->Flush();
 }
@@ -840,20 +697,22 @@ wxLogPassThrough::wxLogPassThrough()
 // static variables
 // ----------------------------------------------------------------------------
 
+bool            wxLog::ms_bRepetCounting = false;
+wxString        wxLog::ms_prevString;
+unsigned int    wxLog::ms_prevCounter = 0;
+time_t          wxLog::ms_prevTimeStamp= 0;
+wxLogLevel      wxLog::ms_prevLevel;
+
 wxLog          *wxLog::ms_pLogger      = (wxLog *)NULL;
-bool            wxLog::ms_doLog        = TRUE;
-bool            wxLog::ms_bAutoCreate  = TRUE;
-bool            wxLog::ms_bVerbose     = FALSE;
+bool            wxLog::ms_doLog        = true;
+bool            wxLog::ms_bAutoCreate  = true;
+bool            wxLog::ms_bVerbose     = false;
 
 wxLogLevel      wxLog::ms_logLevel     = wxLOG_Max;  // log everything by default
 
 size_t          wxLog::ms_suspendCount = 0;
 
-#if wxUSE_GUI
-    const wxChar *wxLog::ms_timestamp  = wxT("%X");  // time only, no date
-#else
-    const wxChar *wxLog::ms_timestamp  = NULL;       // save space
-#endif
+const wxChar   *wxLog::ms_timestamp    = wxT("%X");  // time only, no date
 
 wxTraceMask     wxLog::ms_ulTraceMask  = (wxTraceMask)0;
 wxArrayString   wxLog::ms_aTraceMasks;
@@ -903,12 +762,7 @@ static void wxLogWrap(FILE *f, const char *pszPrefix, const char *psz)
 unsigned long wxSysErrorCode()
 {
 #if defined(__WXMSW__) && !defined(__WXMICROWIN__)
-#ifdef  __WIN32__
     return ::GetLastError();
-#else   //WIN16
-    // TODO what to do on Windows 3.1?
-    return 0;
-#endif  //WIN16/32
 #else   //Unix
     return errno;
 #endif  //Win/Unix
@@ -921,19 +775,33 @@ const wxChar *wxSysErrorMsg(unsigned long nErrCode)
         nErrCode = wxSysErrorCode();
 
 #if defined(__WXMSW__) && !defined(__WXMICROWIN__)
-#ifdef  __WIN32__
-    static wxChar s_szBuf[LOG_BUFFER_SIZE / 2];
+    static wxChar s_szBuf[1024];
 
     // get error message from system
     LPVOID lpMsgBuf;
-    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
-            NULL, nErrCode,
+    if ( ::FormatMessage
+         (
+            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
+            NULL,
+            nErrCode,
             MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
             (LPTSTR)&lpMsgBuf,
-            0, NULL);
+            0,
+            NULL
+         ) == 0 )
+    {
+        // if this happens, something is seriously wrong, so don't use _() here
+        // for safety
+        wxSprintf(s_szBuf, _T("unknown error %lx"), nErrCode);
+        return s_szBuf;
+    }
+
 
     // copy it to our buffer and free memory
-    if( lpMsgBuf != 0 ) {
+    // Crashes on SmartPhone (FIXME)
+#if !defined(__SMARTPHONE__) /* of WinCE */
+    if( lpMsgBuf != 0 )
+    {
         wxStrncpy(s_szBuf, (const wxChar *)lpMsgBuf, WXSIZEOF(s_szBuf) - 1);
         s_szBuf[WXSIZEOF(s_szBuf) - 1] = wxT('\0');
 
@@ -948,26 +816,23 @@ const wxChar *wxSysErrorMsg(unsigned long nErrCode)
                 s_szBuf[len - 2] = wxT('\0');
         }
     }
-    else {
+    else
+#endif // !__SMARTPHONE__
+    {
         s_szBuf[0] = wxT('\0');
     }
 
     return s_szBuf;
-#else   //Win16
-    // TODO
-    return NULL;
-#endif // Win16/32
-#else   // Unix
-#if wxUSE_UNICODE
-    static wxChar s_szBuf[LOG_BUFFER_SIZE / 2];
-    wxConvCurrent->MB2WC(s_szBuf, strerror(nErrCode), WXSIZEOF(s_szBuf) -1);
-    return s_szBuf;
-#else
-    return strerror((int)nErrCode);
-#endif
-#endif  // Win/Unix
+#else // !__WXMSW__
+    #if wxUSE_UNICODE
+        static wchar_t s_wzBuf[1024];
+        wxConvCurrent->MB2WC(s_wzBuf, strerror((int)nErrCode),
+                             WXSIZEOF(s_wzBuf) - 1);
+        return s_wzBuf;
+    #else
+        return strerror((int)nErrCode);
+    #endif
+#endif  // __WXMSW__/!__WXMSW__
 }
 
-#endif //wxUSE_LOG
-
-// vi:sts=4:sw=4:et
+#endif // wxUSE_LOG
